local addonName, a = ...
local L = a.Localize
local s = SpellFlashAddon
local c = BittensGlobalTables.GetTable("BittensSpellFlashLibrary")

local UnitCreatureType = UnitCreatureType
local UnitLevel = UnitLevel
local math = math

local function setWogId(z)
	if s.HasSpell("Eternal Flame") then
		z.ID = c.GetID("Eternal Flame")
	else
		z.ID = c.GetID("Word of Glory")
	end
end

------------------------------------------------------------------------ Common
c.AddOptionalSpell("Blessing of Kings", nil, { NoRangeCheck = 1 })

c.AddOptionalSpell("Blessing of Might", nil, { NoRangeCheck = 1 })

c.AddOptionalSpell("Holy Avenger", nil, { 
	NoGCD = true,
	CheckFirst = function()
		return a.HolyPower < 3
	end
})

c.AddSpell("Judgment", nil, {
	GetDelay = function()
		return a.Judgment
	end,
})

c.AddOptionalSpell("Lay on Hands", nil, {
	FlashColor = "red",
	NoRangeCheck = true,
	CheckFirst = function()
		return s.HealthPercent("player") < 20
			and not s.Debuff(c.GetID("Forebearance"), "player")
	end
})

c.AddInterrupt("Rebuke")

-------------------------------------------------------------------------- Holy
local function hpCheckFirst(z)
	if a.HolyPower < 3 then
		return false
	end
	
	local dp = c.GetBuffDuration("Divine Purpose")
	if dp > 0 and dp < 2.5 then
		z.FlashSize = nil
		z.FlashColor = "red"
		return true
	end
	
	z.FlashColor = "yellow"
	if a.HolyPower < 5 then
		z.FlashSize = s.FlashSizePercent() / 2
	else
		z.FlashSize = nil
	end
	return true
end

c.AddOptionalSpell("Seal of Insight", nil, {
	Type = "form",
})

c.AddOptionalSpell("Word of Glory", "for Holy", {
	RunFirst = setWogId,
	Override = hpCheckFirst,
})

c.AddOptionalSpell("Light of Dawn", "for Holy", {
	Override = hpCheckFirst,
})

c.AddOptionalSpell("Holy Shock", "under 5 with Daybreak", {
	NoRangeCheck = 1,
	CheckFirst = function(z)
		local stack = c.GetBuffStack("Daybreak")
		if stack < 2 and c.IsCasting("Holy Radiance") then
			stack = stack + 1
		end
		c.MakeMini(z, stack == 1)
		return stack > 0 and a.HolyPower < 5
	end
})

c.AddOptionalSpell("Beacon of Light", nil, {
	Override = function()
		return s.InRaidOrParty()
			and not s.MyBuff(c.GetID("Beacon of Light"), a.BeaconTarget)
	end
})

c.AddOptionalSpell("Sacred Shield", "for Holy", {
	NoRangeCheck = true,
	CheckFirst = function()
		return not s.MyBuff(c.GetID("Sacred Shield"), a.SacredShieldTarget)
	end
})

c.AddOptionalSpell("Flash of Light", "under Selfless Healer", { 
	NoRangeCheck = 1,
	CheckFirst = function()
		return c.GetBuffStack("Selfless Healer") == 3
	end
})

c.AddOptionalSpell("Flash of Light", "to Save Selfless Healer", {
	NoRangeCheck = 1,
	FlashColor = "red",
	CheckFirst = function()
		local duration = c.GetBuffDuration("Selfless Healer")
		return duration > 0 and duration < 2.5
	end
})

c.AddOptionalSpell("Judgment", "to Save Selfless Healer", {
	NoRangeCheck = 1,
	FlashColor = "red",
	CheckFirst = function()
		local duration = c.GetBuffDuration("Selfless Healer")
		return duration > 0 and duration < 2.5
	end
})

-------------------------------------------------------------------- Protection
c.AddOptionalSpell("Seal of Insight", "for Prot", {
	Type = "form",
	CheckFirst = function()
		return not s.Form()
	end
})

c.AddOptionalSpell("Righteous Fury", nil, {
	NoGCD = true,
	Buff = "Righteous Fury",
	BuffUnit = "player",
})

c.AddOptionalSpell("Avenging Wrath", "if Plain", {
	NoGCD = true,
	CheckFirst = function()
		return not c.HasTalent("Sanctified Wrath")
			and not c.HasGlyph("Avenging Wrath")
	end
})

c.AddOptionalSpell("Avenging Wrath", "if Cool for Prot", {
	NoGCD = true,
	Enabled = function()
		return c.HasTalent("Sanctified Wrath") or c.HasGlyph("Avenging Wrath")
	end,
	ShouldHold = function()
		return c.HasGlyph("Avenging Wrath") 
			and c.GetHealthPercent("player") > 85
	end
})

c.AddOptionalSpell("Word of Glory", "for Prot", {
	RunFirst = setWogId,
	Override = function(z)
		if c.WearingSet(2, "ProtT15") 
			and c.GetBuffDuration("Shield of Glory") < 3 then
			
			z.FlashColor = "red"
		else
			z.FlashColor = "yellow"
		end
		return a.HolyPower > 0
			and c.GetCooldown(z.ID, true, 1.5) == 0 
			and 95 - c.GetHealthPercent("player")
				> math.min(3, a.HolyPower) 
					* 4 
					* (1 + .1 * c.GetBuffStack("Bastion of Glory", true))
	end
})

c.AddOptionalSpell("Holy Prism", "for Prot", {
	ShouldHold = function()
		return c.GetHealthPercent() < 90
	end,
})

c.AddOptionalSpell("Hand of Purity", nil, {
	NoRangeCheck = true,
})

c.AddOptionalSpell("Divine Protection", nil, {
	NoGCD = true,
	CheckFirst = function()
		return c.HasGlyph("Divine Protection")
	end
})

c.AddOptionalSpell("Light's Hammer", "for Prot", {
	IsUp = function()
		return c.GetCooldown("Light's Hammer", false, 60) > 46
	end,
	ShouldHold = function()
		return c.GetHealthPercent() < 85
	end,
})

c.AddOptionalSpell("Execution Sentence", "for Prot", {
	ShouldHold = function()
		return c.GetHealthPercent() < 85
	end,
})

c.AddOptionalSpell("Holy Avenger", "for Prot", { 
	NoGCD = true,
	ShouldHold = function()
		return a.HolyPower >= 3
	end
})

c.AddOptionalSpell("Guardian of Ancient Kings", nil, {
	NoGCD = true,
	NoRangeCheck = true,
})

c.AddOptionalSpell("Ardent Defender", nil, {
	NoGCD = true,
	Enabled = function()
		return not c.WearingSet(2, "ProtT14")
	end
})

c.AddOptionalSpell("Ardent Defender", "2pT14", {
	NoGCD = true,
	Enabled = function()
		return c.WearingSet(2, "ProtT14")
	end
})

c.AddOptionalSpell("Shield of the Righteous")

c.AddOptionalSpell("Shield of the Righteous", "to save Buffs", {
	NoGCD = true,
	Melee = true,
	Override = function()
		if a.HolyPower < 3 then
			return false
		end
		
		local duration = c.GetBuffDuration("Bastion of Glory")
		if duration > 0 and duration < 1.6 then
			return true
		end
		
		local duration = c.GetBuffDuration("Divine Purpose")
		if duration > 0 and duration < 1.6 then
			return true
		end
	end
})

c.AddSpell("Holy Wrath", nil, {
	Melee = true,
	NoRangeCheck = true,
	NotIfActive = true,
	Cooldown = 9,
})

c.AddSpell("Holy Wrath", "to Stun", {
	Melee = true,
	NotIfActive = true,
	NoRangeCheck = true,
	Cooldown = 9,
	CheckFirst = function()
		local level = UnitLevel("target")
		if level == -1 or level > UnitLevel("player") then
			return false
		end
		
		local ct = UnitCreatureType("target")
		if ct == L["Demon"] or ct == L["Undead"] then
			return true
		end
		
		if not c.HasGlyph("Holy Wrath") then
			return false
		end
		
		return ct == L["Dragonkin"] or ct == L["Elemental"]
	end
})

c.AddSpell("Hammer of the Righteous", "for Debuff", {
	GetDelay = function(z)
		return a.Crusader + c.GetHastedTime(4.5) 
				> c.GetDebuffDuration(c.WEAKENED_BLOWS_DEBUFFS)
			and a.Crusader
	end
})

c.AddSpell("Sacred Shield", "for Prot", {
	Buff = "Sacred Shield",
	BuffUnit = "player",
	NoRangeCheck = true,
	Cooldown = 6,
})

c.AddSpell("Sacred Shield", "Refresh", {
	NoRangeCheck = true,
	GetDelay = function()
		return math.max(
			c.GetBuffDuration("Sacred Shield") - 5, 
			c.GetCooldown("Sacred Shield", false, 6))
	end
})

c.AddSpell("Flash of Light", "for Prot", {
	NoRangeCheck = true,
	CheckFirst = function()
		return a.SelflessHealer == 3 and c.GetHealthPercent("player") < 85
	end
})

c.AddSpell("Crusader Strike", "for Prot", {
	GetDelay = function()
		return a.Crusader
	end,
})

c.AddSpell("Crusader Strike", "Delay", {
	IsMinDelayDefinition = true,
	GetDelay = function()
		return a.Crusader, .2
	end,
})

c.AddSpell("Avenger's Shield", nil, {
	Cooldown = 15,
})

c.AddSpell("Avenger's Shield", "under Grand Crusader", {
	Cooldown = 15,
	CheckFirst = function()
		return c.HasBuff("Grand Crusader")
	end
})

c.AddSpell("HP Gen Delay for Prot", nil, {
	ID = "Crusader Strike",
	IsMinDelayDefinition = true,
	GetDelay = function()
		return math.min(a.Judgment, a.Crusader), c.LastGCD - .2
	end
})

c.AddSpell("Hammer of Wrath", nil, {
	Cooldown = 6,
	CheckFirst = function()
		return a.HoWPhase
	end
})

c.AddOptionalSpell("Consecration", nil, {
	NoRangeCheck = true,
	GetDelay = function(z)
		if c.HasGlyph("Consecration") then
			z.Melee = nil
			return c.GetCooldown("Consecration Glyphed", false, 9)
		else
			z.Melee = true
			return c.GetCooldown("Consecration", false, 9)
		end
	end
})

c.AddTaunt("Hand of Reckoning", nil, { NoGCD = true })

------------------------------------------------------------------- Retribution
c.AddOptionalSpell("Seal of Truth", nil, {
	Type = "form",
	CheckFirst = function()
		return s.Form() == nil
	end
})

c.AddOptionalSpell("Word of Glory", "for Ret", {
	RunFirst = setWogId,
	Override = function()
		return a.HolyPower > 0
			and not s.InRaidOrParty() 
			and 95 - c.GetHealthPercent("player") 
				> 8 * math.min(3, a.HolyPower)
			and (not s.InCombat() or a.HolyPower >= 3)
	end
})

c.AddOptionalSpell("Flash of Light", "for Ret", {
	NoRangeCheck = true,
	CheckFirst = function()
		return (a.SelflessHealer == 3 or not s.InCombat())
			and c.GetHealthPercent("player") < 80
	end
})

c.AddOptionalSpell("Avenging Wrath", "for Ret", { 
	CheckFirst = function()
		return a.Inquisition > 0
	end
})

c.AddSpell("Inquisition", nil, {
	GetDelay = function()
		return a.HolyPower >= 1 and a.Inquisition
	end
})

c.AddSpell("Inquisition", "before Templar's Verdict at 5", {
	GetDelay = function()
		return (a.HolyPower == 5 
				or (a.HolyPower >= 3 and a.HolyAvenger > 0))
			and a.Inquisition - 3
	end
})

c.AddSpell("Inquisition", "before Templar's Verdict", {
	GetDelay = function()
		return a.HolyPower >= 3 and a.Inquisition - 4
	end
})

c.AddOptionalSpell("Execution Sentence", nil, {
	Cooldown = 60,
	CheckFirst = function()
		return a.Inquisition > 0
	end
})

c.AddSpell("Templar's Verdict", nil, { 
	Melee = true,
	CheckFirst = function()
		return a.HolyPower >= 3
	end
})

c.AddSpell("Templar's Verdict", "at 5", { 
	Melee = true,
	CheckFirst = function()
		return (a.HolyPower == 5 
				or (a.HolyPower >= 3 and a.HolyAvenger > c.LastGCD))
	end,
})

c.AddSpell("Templar's Verdict", "4pT15", { 
	Melee = true,
	CheckFirst = function()
		return a.HolyPower >= 3 
			and a.Crusader < c.LastGCD
			and c.HasBuff("Ret 4pT15 Buff")
	end,
})

c.AddSpell("Hammer of Wrath", "for Ret", {
	GetDelay = function(z)
		local cd = c.GetCooldown("Hammer of Wrath", false, 6)
		z.WhiteFlashOffset = cd
		return (a.HoWPhase or a.AvengingWrath > 0)
			and s.SpellInRange(c.GetID("Hammer of Wrath"))
			and cd - .2
	end,
})

c.AddSpell("Exorcism", nil, {
	GetDelay = function()
		return (not c.HasGlyph("Mass Exorcism") or s.MeleeDistance())
			and a.Exorcism
	end,
})

local function countCasts(span)
	local casts = 0
	if a.Exorcism < span then
		casts = casts + 1
	end
	
	local hammer = c.GetCooldown("Hammer of Wrath", false, 6)
	if hammer < span and (a.HoWPhase or a.AvengingWrath >= hammer) then
		casts = casts + 1
	end
	
	if c.GetCooldown("Execution Sentence", false, 60) < span then
		casts = casts + 1
	end
	
	if a.HolyAvenger > 0 then
		casts = casts + 1
		if a.HolyAvenger > 2 * c.LastGCD then
			casts = casts + 1
		end
	elseif a.RawHolyPower >= 3 then
		casts = casts + 1
		if c.HasBuff("Divine Purpose") then
			casts = casts + 1
		end
	end
	
	return casts
end

c.AddSpell("Judgment", "unless Wastes GCD", {
	Cooldown = 6,
	CheckFirst = function()
		return (c.HasTalent("Sanctified Wrath") and a.AvengingWrath > 0)
			or (countCasts(2 * c.LastGCD) > 0 and countCasts(3 * c.LastGCD) > 1)
	end
})

c.AddSpell("Crusader Strike", "for Ret", {
	GetDelay = function()
		return a.Crusader
	end,
})

c.AddSpell("Crusader Strike", "4pT15", {
	GetDelay = function(z)
		z.WhiteFlashOffset = a.Crusader
		return c.WearingSet(4, "RetT15") 
			and c.GetBuffDuration("Ret 4pT15 Buff") < a.Crusader + c.LastGCD
			and a.Crusader - .2
	end
})

c.AddSpell("HP Gen Delay for Ret", nil, {
	ID = "Crusader Strike",
	IsMinDelayDefinition = true,
	GetDelay = function()
		return math.min(a.Exorcism, a.Judgment, a.Crusader), .2
	end
})

c.AddOptionalSpell("Light's Hammer", nil, {
	Cooldown = 60,
})

c.AddSpell("Holy Prism", nil, {
	Cooldown = 20,
})

c.AddOptionalSpell("Sacred Shield", "for Ret", {
	GetDelay = function(z)
		local delay = math.max(
			c.LastGCD, c.GetCooldown("Sacred Shield", false, 6))
		z.WhiteFlashOffset = -delay
		return delay
	end
})
